/** \file alleggl.h
  * \brief Main header file for AllegroGL
  */

#ifndef _allegrogl_included_alleggl_h
#define _allegrogl_included_alleggl_h

#include <allegro.h>
#ifdef _WIN32
#include <winalleg.h>
#endif

#include "agl_ext.h"

#define _inline

#ifdef AGL_USE_EXTENSIONS
#ifdef AGL_SOURCE_COMPILE
#include "gl.h"
#include "glprocs.h"
#ifdef ALLEGRO_WINDOWS
#include "wglext.h"
#endif
#else
#include <GLext/gl.h>
#include <GLext/glprocs.h>
#ifdef ALLEGRO_WINDOWS
#include <GLext/wglext.h>
#endif
#endif
#else
#include <GL/gl.h>
#endif
#undef _inline

#ifdef ALLEGRO_WITH_XWINDOWS
#ifndef HAVE_LIBPTHREAD
   #error AllegroGL requires Allegro to have pthread support enabled!
#endif
#include <alleggl_config.h>
#endif


/** \defgroup version Version information
  * \{ */
/** \name Version Information
  * \{ */
#define AGL_VERSION 0                ///< Major version number
#define AGL_SUB_VERSION 1            ///< Minor version number
#define AGL_WIP_VERSION 4            ///< Work-In-Progress version number
#define AGL_VERSION_STR "0.1.4"      ///< Version description string
/** \} */
/** \} */

/* Version Check */
#if (ALLEGRO_VERSION <= 3)
	#error AllegroGL requires Allegro 4.0.0 or later to compile!
#endif
#ifndef GL_VERSION_1_1
	#error AllegroGL requires OpenGL 1.1 libraries or later to compile!
#endif


#ifdef __cplusplus
extern "C" {
#endif

/* Missing #defines from Mingw */
#ifndef PFD_SWAP_LAYER_BUFFERS
#define PFD_SWAP_LAYER_BUFFERS  0x00000800
#endif

#ifndef PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED 0x00001000
#endif

#ifndef PFD_SUPPORT_DIRECTDRAW
#define PFD_SUPPORT_DIRECTDRAW  0x00002000
#endif

#ifndef CDS_FULLSCREEN
#define CDS_FULLSCREEN      0x00000004
#endif

#ifndef ENUM_CURRENT_SETTINGS
#define ENUM_CURRENT_SETTINGS       ((DWORD)-1)
#endif


#define AGL_ERROR_SIZE 256
extern char allegro_gl_error[];


struct allegro_gl_driver {
	void (*flip) (void);
	void (*gl_on) (void);
	void (*gl_off) (void);
};


/** \defgroup core Core routines
 *  Basic AllegroGL functions. These functions allow you to initialize
 *  AllegroGL, set up a rendering context via set_gfx_mode(), and allow
 *  access to regular OpenGL functions.
 *
 */
/** \name Core Functions
 *  \{
 */
int install_allegro_gl(void);
void remove_allegro_gl(void);

void allegro_gl_flip (void);
float allegro_gl_opengl_version();

/** \ingroup core
 *  Backward compatibility #define for programs written
 *  prior to AGL 0.0.22.
 *  It isn't defined as anything meaningful, so you don't
 *  need to put them in your program.
 *
 * \see allegro_gl_end()
 */
#define allegro_gl_begin() ;

/** \ingroup core
 *  Backward compatibility #define for programs written
 *  prior to AGL 0.0.22.
 *  It isn't defined as anything meaningful, so you don't
 *  need to put them in your program.
 *
 * \see allegro_gl_begin()
 */
#define allegro_gl_end() ;

/** \} */


struct allegro_gl_rgba_size {
	int r, g, b, a;
};
struct allegro_gl_indexed_size {
	int bpp;
};
union allegro_gl_pixel_size {
	struct allegro_gl_rgba_size rgba;
	struct allegro_gl_indexed_size indexed;
};

struct allegro_gl_display_info {
	int allegro_format;
	union allegro_gl_pixel_size pixel_size;
	int colour_depth;
	union allegro_gl_pixel_size accum_size;
	int doublebuffered;
	int stereo;
	int aux_buffers;
	int depth_size;
	int stencil_size;
	int w, h, x, y;
	int r_shift, g_shift, b_shift, a_shift;
	int packed_pixel_type, packed_pixel_format;
	int rmethod; /* type of rendering, 0: software; 1: hardware, 2: unknown (X Win) */
	int fullscreen;
	int mesa_driver;
};


extern struct allegro_gl_display_info allegro_gl_display_info;


/** \addtogroup settings 
 *  \{ */
/** \name Frame Buffer Settings
 *  \{ */

#define AGL_ALLEGRO_FORMAT  0x00000001 ///< Use Allegro-compatible framebuffer (depricated)

/** Select the red depth of the frame buffer. This defines the number of
 *  bits for the red component you'd like. The driver may or may not find
 *  a suitable mode
 */
#define AGL_RED_DEPTH       0x00000002

/** Select the green depth of the frame buffer. This defines the number of
 *  bits for the green component you'd like. The driver may or may not find
 *  a suitable mode
 */
#define AGL_GREEN_DEPTH     0x00000004

/** Select the blue depth of the frame buffer. This defines the number of
 *  bits for the blue component you'd like. The driver may or may not find
 *  a suitable mode
 */
#define AGL_BLUE_DEPTH      0x00000008

/** Select the alpha depth of the frame buffer. This defines the number of
 *  bits for the alpha component you'd like. Not many video cards support
 *  destination alpha, so be aware that the you may not get an alpha depth
 *  at all.
 */
#define AGL_ALPHA_DEPTH     0x00000010

/** Specify the total color depth of the frame buffer. The driver is free to
 *  select any combination of red, green, blue and alpha bits.
 */
#define AGL_COLOR_DEPTH     0x00000020


/** Select the red depth of the accumulator buffer. This defines the number
 *  of bits to use for the red component. The driver may or may not find a
 *  suitable mode. Note that on many video cards, the accumulator buffer is
 *  not accelerated.
 */
#define AGL_ACC_RED_DEPTH   0x00000040 ///< Set the Red color depth (in bits) of the accumulation buffer

/** Select the green depth of the accumulator buffer. This defines the number
 *  of bits to use for the green component. The driver may or may not find a
 *  suitable mode. Note that on many video cards, the accumulator buffer is
 *  not accelerated.
 */
#define AGL_ACC_GREEN_DEPTH 0x00000080 ///< Set the Green color depth (in bits) of the accumulation buffer


/** Select the blue depth of the accumulator buffer. This defines the number
 *  of bits to use for the blue component. The driver may or may not find a
 *  suitable mode. Note that on many video cards, the accumulator buffer is
 *  not accelerated.
 */
#define AGL_ACC_BLUE_DEPTH  0x00000100 ///< Set the Blue color depth (in bits) of the accumulation buffer


/** Select the alpha depth of the accumulator buffer. This defines the number
 *  of bits to use for the alpha component. The driver may or may not find a
 *  suitable mode. Note that on many video cards, the accumulator buffer is
 *  not accelerated. Not many cards support destination alpha either.
 */
#define AGL_ACC_ALPHA_DEPTH 0x00000200 ///< Set the Alpha color depth (in bits) of the accumulation buffer

/** Creates a backbuffer if set. The buffering mode may be double buffering
  * or page flipping, depending on the driver settings. OpenGL programs cannot
  * chose the buffering mode themselves.
  */
#define AGL_DOUBLEBUFFER    0x00000400

/** Creates seperate left/right buffers for stereo display. Stereo display
 *  is used with special hardware (tipically glasses) for giving the illusion
 *  of depth by drawing the left and right buffers with a slight horizontal
 *  displacement. This makes the display appear to heavily flicker without
 *  the special hardware.
 *  Set to TRUE for enabling stereo viewing.
 */
#define AGL_STEREO          0x00000800


/** Creates additional auxiliary buffers. This allows you to have more than
 *  one rendering context. Few video cards support this feature.
 */
#define AGL_AUX_BUFFERS     0x00001000


/** Select the depth of the z-buffer. The z-buffer is used to properly
 *  display polygons in 3D without recurring to sorting. The higher the
 *  depth of the z-buffer, the more precise it is.
 */
#define AGL_Z_DEPTH         0x00002000


/** Select the depth of the stencil buffer. The stencil buffer is used to
 *  to do per-pixel testing (like the z-buffer), but of an arbitrary
 *  pattern instead of depth. Please see the OpenGL documentation for details.
 *  Newer cards support stenciling in hardware, but older cards (TNT2s,
 *  Voodoos, ATI Rage) do not.
 */
#define AGL_STENCIL_DEPTH   0x00004000

/** Requests a placement of the window to a specified pixel location.
 *  The driver may or may not honnor the request.
 */
#define AGL_WINDOW_X        0x00008000

/** Same as AGL_WINDOW_X, but for the y-axis.
 */
#define AGL_WINDOW_Y        0x00010000

/** Set it if you'd like AllegroGL to pay special attention on whether
 *  hardware acceleration is present or not. Notice however this this
 *  isn't a garentee that the OpenGL operations will be accelerated, but
 *  rather a request that the operations go through the video card's
 *  drivers instead of Microsoft's software renderer. The video card
 *  driver is free to drop back to software mode if it deems it necessary.
 *  This setting has no effect in X.
 */
#define AGL_RENDERMETHOD    0x00020000

/** Set if you'd like a full screen mode. Full screen may or may not be
 *  available on the current platform.
 */
#define AGL_FULLSCREEN      0x00040000


/** Set if you'd like a windowed mode. Windowed modes may or may not be
 *  available on the current platform.
 */
#define AGL_WINDOWED        0x00080000
/** \} */


/** \name Mode slection priority options
 *  \{ */
#define AGL_DONTCARE   0 ///< Ignore these settings
#define AGL_SUGGEST   -1 ///< Prefer the assigned values for these settings
#define AGL_REQUIRE   -2 ///< Reject other values for these settings
/** \} */


/** \name Mode selection functions
 *  \{ */
void allegro_gl_clear_settings (void);
void allegro_gl_set (int option, int value);
int allegro_gl_get (int option);
void allegro_gl_save_settings();
void allegro_gl_load_settings();
/** \} */
/** \} */


/** \defgroup gfxdrv Graphics drivers
  *
  * Use set_gfx_mode to select an OpenGL mode as normal, but using e.g.
  * GFX_OPENGL as the driver.  The virtual width and height are ignored.
  * To set the colour depth, use allegro_gl_set (AGL_COLOR_DEPTH, xxx).
  * However if the color depth is not set by allegro_gl_set(), AllegroGL
  * will refer to the value set by the last call to set_color_depth().
  *
  * Allegro modes are still available. Use of GFX_AUTODETECT or
  * GFX_AUTODETECT_WINDOWED will select Allegro modes, and not OpenGL modes.
  */

#if defined DOXYGEN  /* Is this a documentation scan? */
  /// \ingroup gfxdrv
  /// \name Graphics drivers
  /// \{
  /** Windowed OpenGL graphics driver for Allegro */
  #define GFX_OPENGL_WINDOWED
  /** Fullscreen OpenGL graphics driver for Allegro */
  #define GFX_OPENGL_FULLSCREEN
  /** Non-specific OpenGL graphics driver for Allegro */
  #define GFX_OPENGL
  /// \}

#else
#if defined ALLEGROGL_GENERIC_DRIVER
  /* Allegro is able to determine at run-time if windowed or fullscreen modes
     are available */
  #define GFX_OPENGL_WINDOWED               AL_ID('O','G','L','W')
  #define GFX_OPENGL_FULLSCREEN             AL_ID('O','G','L','F')

#else
#if defined _WIN32
  /* Windows always supports fullscreen */
  #define GFX_OPENGL_WINDOWED               AL_ID('O','G','L','W')
  #define GFX_OPENGL_FULLSCREEN             AL_ID('O','G','L','F')

#elif defined ALLEGROGL_HAVE_XF86VIDMODE
  /* X fullscreen support only works with the XF86 VidMode extension */
  #define GFX_OPENGL_WINDOWED               AL_ID('O','G','L','W')
  #define GFX_OPENGL_FULLSCREEN             AL_ID('O','G','L','F')

#elif defined ALLEGRO_WITH_XWINDOWS
  /* Plain X only supports windowed modes */
  #define GFX_OPENGL_WINDOWED               AL_ID('O','G','L','W')
  
#else
  #warn Unknown or unsupported platform.
#endif
#endif

#define GFX_OPENGL      		    AL_ID('O','G','L','D')

#endif


/* Declare graphics driver objects */
extern GFX_DRIVER gfx_allegro_gl_default;
#ifdef GFX_OPENGL_WINDOWED
extern GFX_DRIVER gfx_allegro_gl_windowed;
#endif
#ifdef GFX_OPENGL_FULLSCREEN
extern GFX_DRIVER gfx_allegro_gl_fullscreen;
#endif



extern BITMAP *allegro_gl_screen;
extern BITMAP *__allegro_gl_memory;


/// \name Screen update modes
/// \{
/** Default mode, and fastest for OpenGL operations. */
#define AGL_MODE_DIRECT     0
#define AGL_MODE_DUPLEX     1
#define AGL_MODE_OPAQUE     2
#define AGL_MODE_MASKED     3
#define AGL_MODE_OPAQUE_RT  4
#define AGL_MODE_MASKED_RT  5
/// \}


/** \defgroup texture Texture Routines
 *  AllegroGL provides functions to allow using Allegro BITMAP objects to
 *  be used as OpenGL textures. There are however several limitations.
 *  8 bpp BITMAPs are not fully supported. 15 and 16 bpp BITMAPs are
 *  not directly supported in Windows, so AllegroGL will convert those
 *  to 24 bpp before sending them to the OpenGL driver.
 */
/** \{ */
/** \name Texture routines
 *  \{ */
GLint allegro_gl_get_texture_format(BITMAP *bmp);
GLint allegro_gl_set_texture_format(GLint format);
GLenum allegro_gl_bitmap_type(BITMAP *bmp);
GLenum allegro_gl_bitmap_color_format(BITMAP *bmp);
int allegro_gl_use_mipmapping(int enable);
int allegro_gl_use_alpha_channel(int enable);
int allegro_gl_check_texture (BITMAP *bmp);
GLuint allegro_gl_make_texture (BITMAP *bmp);
GLuint allegro_gl_make_masked_texture (BITMAP *bmp);
/** \} */
/** \} */



/** \addtogroup allegro Allegro Interfacing
 *
 * \{ */
/** \name Allegro Interfacing routines
 *  \{ */
int allegro_gl_screen_mode (int newmode);
void allegro_gl_set_allegro_mode (void);
void allegro_gl_unset_allegro_mode (void);
void allegro_gl_set_projection (void);
void allegro_gl_unset_projection (void);
/** \} */
/** \} */



/** \defgroup math Math conversion routines
  * Routines to convert between OpenGL and Allegro math types
  */
/** \{ */

/** \name Matrix conversion routines
 *  \{ */

void allegro_gl_MATRIX_to_GLfloat (MATRIX *m, GLfloat gl[16]);
void allegro_gl_MATRIX_to_GLdouble (MATRIX *m, GLdouble gl[16]);
void allegro_gl_MATRIX_f_to_GLfloat (MATRIX_f *m, GLfloat gl[16]);
void allegro_gl_MATRIX_f_to_GLdouble (MATRIX_f *m, GLdouble gl[16]);

void allegro_gl_GLfloat_to_MATRIX (GLfloat gl[16], MATRIX *m);
void allegro_gl_GLdouble_to_MATRIX (GLdouble gl[16], MATRIX *m);
void allegro_gl_GLfloat_to_MATRIX_f (GLfloat gl[16], MATRIX_f *m);
void allegro_gl_GLdouble_to_MATRIX_f (GLdouble gl[16], MATRIX_f *m);

/** \} */

/** \name Quaternion conversion routines
 *  \{ */
void allegro_gl_apply_quat(QUAT *q);
void allegro_gl_quat_to_glrotatef(QUAT *q, float *angle, float *x, float *y, float *z);
void allegro_gl_quat_to_glrotated(QUAT *q, double *angle, double *x, double *y, double *z);
/** \} */
/** \} */


/** \defgroup Text Text drawing and fonts.
 *
 *  AllegroGL now provides mechanisms for converting Allegro FONTs to
 *  a format which is more usable in OpenGL. You can also load system
 *  fonts (such as Arial and Courrier) and draw using those.
 *  Allegro FONTs should still work if you use textout, textprintf, etc
 *  on them. However, converted Allegro FONTs will not work with these
 *  functions. You will have to use AllegroGL functions to work with
 *  AllegroGL FONTs.
 *
 * \{ */

/** \name Text Drawing and Font conversion
 * \{
 */


/* These define the supported font types */
/** Indicates that you don't really care how a font will be converted.
 *  AGL will pick the best format for you.
 */
#define AGL_FONT_TYPE_DONT_CARE     -1

/** Indicates that you want fonts to be converted to a bitmap format.
 *  Bitmaps are used to represent characters with one bit per pixel,
 *  and are usually very slow to draw, as they stall the 3D pipeline.
 *  Note that you can't scale, rotate or place text with a z coordinate
 *  if you use this style.
 *  A display list will be created.
 *
 *  \sa allegro_gl_convert_allegro_font(), allegro_gl_printf()
 */
#define AGL_FONT_TYPE_BITMAP         0

/** Indicates that you want fonts to be converted to an outline format.
 *  Outlined mode is a vector-style font. Characters are represented by a
 *  set of polygons and lines. This style of fonts is fast to draw, and can
 *  be scaled, rotated, etc, since they are just vectors.
 *  A display list will be created.
 *
 *  \sa allegro_gl_convert_allegro_font(), allegro_gl_printf()
 */
#define AGL_FONT_TYPE_OUTLINE        1

/** Indicates that you want fonts to be converted to a texture format.
 *  Each character is represented by a textured quad. The texture is
 *  common for all characters of the same font, and it will automatically be
 *  uploaded to the video card when needed. Drawing text with this
 *  type of font is the fastest since only two triangles are needed per
 *  character. Textured text can also be scaled and rotated.
 *  A display list will be created.
 *
 *  \sa allegro_gl_convert_allegro_font(), allegro_gl_printf()
 */
#define AGL_FONT_TYPE_TEXTURED       2


/* These defines the font styles for system generated fonts */
/** Creates a font with bold characters. System fonts only. */
#define AGL_FONT_STYLE_BOLD          1

/** Creates a font with black (strong bold) characters. System fonts only. */
#define AGL_FONT_STYLE_BLACK         2

/** Creates a font with italicized characters. System fonts only. */
#define AGL_FONT_STYLE_ITALIC        4

/** Creates a font with underlined characters. System fonts only. */
#define AGL_FONT_STYLE_UNDERLINE     8

/** Creates a font with striked out characters. System fonts only. */
#define AGL_FONT_STYLE_STRIKEOUT    16

/** Creates a font with anti-aliased characters. System fonts only.
 * Anti-aliasing may not be available, and no error will be reported
 * if such is the case. 
 */
#define AGL_FONT_STYLE_ANTI_ALIASED 32

/** Font generation mode. System fonts only.
 * Indicates that you want outline system fonts to be generated using polygons.
 *
 * \sa allegro_gl_load_system_font(), allegro_gl_load_system_font_ex()
 */
#define AGL_FONT_POLYGONS 1


/** Font generation mode. System fonts only.
 * Indicates that you want outline system fonts to be generated using lines.
 *
 * \sa allegro_gl_load_system_font(), allegro_gl_load_system_font_ex()
 */
#define AGL_FONT_LINES    2

/** */

extern int allegro_gl_printf(FONT *f, float x, float y, float z, int color, char *format, ...);
extern FONT *allegro_gl_convert_allegro_font(FONT *f, int type, float scale);
extern void allegro_gl_set_font_generation_mode(int mode);
extern FONT *allegro_gl_load_system_font(char *name, int style, int w, int h);
extern FONT *allegro_gl_load_system_font_ex(char *name, int type, int style, int w, int h, float depth, int start, int end);
extern void allegro_gl_destroy_font(FONT *f);

/** \} */
/** \} */


/** \defgroup extensions OpenGL Extensions
 * Management of the OpenGL extensions mechanism.
 * AllegroGL provides two ways to access OpenGL extensions : Intel's GLsdk
 * library and some versatile portable routines.
 *
 * Based on the extension headers, Intel has created the GLsdk library
 * to ease use of OpenGL extensions, by providing a static library containing
 * wrapper functions for extensions which otherwise must be accessed by use of
 * \b wglGetProcAddress on Windows and \b glXGetProcAddressARB on Unix.
 * Sources can be downloaded at http://oss.sgi.com/projects/ogl-sample/sdk.html
 * See readme.txt for more informations on how to compile GLsdk in order to use
 * it with AllegroGL.
 *
 * If you want to get more control on extensions or if you want to use an
 * extension that is not supported by GLsdk then you can use AllegroGL's own
 * routines : allegro_gl_is_extension_supported() and
 * allegro_gl_get_proc_address() provide a way to determine if an extension is
 * available and to get its address. These routines are available on every
 * platforms that AllegroGL supports
 *
 * \b Important \b Note : Functions activated by the GLsdk library or
 * allegro_gl_get_proc_address() are only defined for the AllegroGL's context.
 * If you create another GL context then you must get the functions addresses
 * by some additional calls to allegro_gl_get_proc_address().
 */
/** \{ */
/** \name OpenGL Extensions Management Functions
 *  \{
 */
#if defined DOXYGEN  /* Is this a documentation scan? */
/** OpenGL extensions handlers helper.
 * Defines a function pointer type. This macro is almost equivalent to a
 * \b typedef. It is intended to hide some platform-specific machinery in
 * order to keep code portable.
 *
 * \sa allegro_gl_get_proc_address()
 */
#define AGL_DEFINE_PROC_TYPE

#else

int allegro_gl_is_extension_supported(const char *);
void *allegro_gl_get_proc_address(const char *);

#if defined ALLEGRO_WINDOWS
	#define AGL_DEFINE_PROC_TYPE(type, name, args) \
		type (APIENTRY * name) args;
#else
	#define AGL_DEFINE_PROC_TYPE(type, name, args) \
		type (*name) args;
#endif

#endif
/** \} */
/** \} */


/** \defgroup gui Allegro-compatible GUI routines
 *  AllegroGL GUI wrappers.
 *  Due to several specificities of OpenGL, some of the Allegro's GUI routines
 *  can not be used "as is". Indeed they are not designed to natively support
 *  double-buffered graphics mode. Hence AllegroGL provides wrapper routines
 *  of do_dialog, alert and so on...
 *
 *  AllegroGL GUI routines internally call allegro_gl_set_allegro_mode() and its
 *  counterpart allegro_gl_unset_allegro_mode(). So the default drawing mode in
 *  the GUI routines is the "2D Allegro mode" and functions like line() or
 *  rect() can safely be called.
 *
 *  Additionnaly AllegroGL provides a new GUI object d_algl_viewport_proc()
 *  which allows to have a 3D viewport that can safely coexist with other
 *  "classical" 2D GUI objects : no need to call allegro_gl_set_allegro_mode()
 *  or to save the current state of OpenGL.
 */
/** \{ */
int algl_do_dialog (DIALOG *dialog, int focus_obj);
int algl_popup_dialog (DIALOG *dialog, int focus_obj);
void algl_draw_mouse (void);
void algl_set_mouse_drawer (void (*user_draw_mouse)(void));
int algl_alert(AL_CONST char *s1, AL_CONST char *s2, AL_CONST char *s3, AL_CONST char *b1, AL_CONST char *b2, int c1, int c2);
int algl_alert3(AL_CONST char *s1, AL_CONST char *s2, AL_CONST char *s3, AL_CONST char *b1, AL_CONST char *b2, AL_CONST char *b3, int c1, int c2, int c3);
int d_algl_viewport_proc(int msg, DIALOG *d, int c);
/** \} */



#ifdef __cplusplus
}
#endif


/* Fixes to MS's (or SGI?) broken GL headers */
#ifdef GL_VERSION_1_1
#ifndef GL_TEXTURE_BINDING_2D

#ifdef GL_TEXTURE_2D_BINDING
#define GL_TEXTURE_BINDING_2D GL_TEXTURE_2D_BINDING
#endif

#else

#ifdef GL_TEXTURE_BINDING_2D
#define GL_TEXTURE_2D_BINDING GL_TEXTURE_BINDING_2D
#endif

#endif

#ifndef GL_TEXTURE_BINDING_2D
#warning "GL_TEXTURE_BINDING_2D or GL_TEXTURE_2D_BINDING isn't defined by your"
#warning OpenGL headers. Make sure you have a genuine set of headers for
#warning OpenGL 1.1 (or greater)
#endif
#endif

#endif
